
local configuration = {}

local safer = require("safer")
local schema = require("schema")

local function get(config_data, filename, config)
   local config = config or {}
   local chunk, err = load(config_data, filename, "t", config)
   if not chunk then
      return nil, "Failed loading config file "..filename..": "..err
   end
   config.os = { getenv = os.getenv }
   config.tonumber = tonumber

   local ok, err = pcall(chunk)
   config.os = nil
   if not ok then
      return nil, "Failed processing config file "..filename..": "..err
   end
   return config
end

local function get_from_multiple(filename, sga_name, config)
   local err = "Failed getting config from multiple config file %s: %s"
   if not sga_name then
      return nil, string.format(err, filename, "missing sga_name argument")
   end
   local config_data = config.sga[sga_name]
   if not config_data then
      return nil, string.format(err, filename, "'"..sga_name.."' not found")
   end
   config.sga = nil -- avoid multiple inside multiple
   
   config, err = get(config_data, filename, config)
   if not config then
      return nil, string.format("[%s] %s", sga_name, err)
   end
   if config.sga_name then
      return nil, string.format("Duplicate definition of sga_name: '%s' (from file) & '%s' (from argument). Multiple config file must not define sga_name value.", config.sga_name, sga_name)
   end

   config.sga_name = sga_name

   return config
end

function configuration.read(filename, sga_name)
   local config_fd = io.open(filename, "r")
   if not config_fd then
      return nil, "Could not read config file "..filename
   end
   local config_data = config_fd:read("*a")
   config_fd:close()

   local config, err = {}
   config.sga = {}
   config, err = get(config_data, filename, config)
   if not config then
      return nil, err
   end

   local is_multiple = (config.sga and next(config.sga))
   if sga_name and not is_multiple then
      return nil, string.format(
         "Failed getting config '%s': file %s is not a valid multiple config file",
         sga_name, filename)
   end
   if is_multiple then
      config, err = get_from_multiple(filename, sga_name, config)
      if not config then
         return nil, err
      end
   end
   
   config = safer.readonly(config)
   return config
end

function configuration.check(config)
   local config_schema = schema.Record {
      csbase_server = schema.Pattern("https?://.*"),
      platform = schema.String,
      sgad_host = schema.String,
      sgad_port = schema.NumberFrom(1,65535),
      sga_name = schema.Pattern("[A-Za-z0-9_]+"),
      sga_type = schema.Optional(schema.String),
      status_interval_s = schema.PositiveNumber,
      exec_polling_interval_s = schema.PositiveNumber,
      driver = schema.Test(function(v)
         local ok, err = pcall(require, v)
         if not ok then
            io.stderr:write("Failed loading driver '"..tostring(v).."': "..err.."\n")
         end
         return ok
      end, "must be a valid module"),
      driver_config = schema.Optional(schema.Map(schema.String, schema.Any)),
      resources = schema.Optional(schema.Collection(schema.String)),
      extra_config = schema.Optional(schema.Map(schema.String, schema.Any)),
   }
   local err = schema.CheckSchema(config, config_schema)
   if err then
      local prefix = ""
      if config.sga_name then
         prefix =  "["..config.sga_name.."] "
      end
      io.stderr:write(prefix.."Configuration error: "..tostring(err).."\n")
      os.exit(1)
   end
end

return configuration
